import json

from django.db.models import Q

from django_spire.core.filtering.constants import NOT_SELECTED_CHOICES


# Used for boolean fields filtered with select field.
def query_bool_select_field(
        query: Q,
        field_name: str,
        field_value: str | None
) -> Q:
    if field_value in NOT_SELECTED_CHOICES:
        return query

    if isinstance(field_value, str):
        field_value = json.loads(field_value)

    query &= Q(**{field_name: bool(field_value)})

    return query


def query_multi_select_field(
        query: Q,
        field_name: str,
        field_value: list[str] | str
) -> Q:
    if not field_value:
        return query

    if isinstance(field_value, str):
        field_value = json.loads(field_value)

    field_values = [
        value for value in field_value
        if value not in NOT_SELECTED_CHOICES
    ]

    if field_values:
        query &= Q(**{field_name + '__in': field_values})

    return query


def query_select_field(
        query: Q,
        field_name: str,
        field_value: str | None,
) -> Q:
    if field_value not in NOT_SELECTED_CHOICES:
        query &= Q(**{field_name: field_value})

    return query



from __future__ import annotations

import json

from typing_extensions import TYPE_CHECKING

from django.template.response import TemplateResponse

from django_spire.notification.app.models import AppNotification

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def notification_dropdown_template_view(request: WSGIRequest) -> TemplateResponse:
    app_notification_list = (
        AppNotification.objects.active()
        .is_sent()
        .annotate_is_viewed_by_user(request.user)
        .select_related('notification')
        .distinct()
        .order_by('-notification__sent_datetime')
        .ordered_by_priority()
    )

    body_data = json.loads(request.body.decode('utf-8'))

    return TemplateResponse(
        request,
        context={
            'app_notification_list': app_notification_list,
            'app_notification_list_url': body_data.get('app_notification_list_url'),
        },
        template='django_spire/notification/app/dropdown/notification_dropdown_content.html'
    )



from django_spire.core.filtering.querysets import FilterQuerySet, SearchQuerySet
from django_spire.core.filtering.session import QuerySetFilterSession


class QuerySetFilter:
    filter_key: str = None

    def __init__(
            self,
            request,
            filter_key: str,
            default_filtering_data: dict = None
    ):
        self.request = request
        self.filter_key = filter_key

        self.user_command = self.request.GET.get('filter_type')
        self.user_key = self.request.GET.get('filter_key')

        self.session_helper = QuerySetFilterSession(self.request, self.filter_key)

        if not self.request.GET.get('page') and self.user_command not in ['filter', 'search']:
            self.session_helper.clear()

        if self.user_command == 'clear':
            self.session_helper.clear()
            self.filter_data = default_filtering_data
        elif self.request.GET and self.filter_key == self.user_key:
            self.filter_data = self.request.GET
        elif self.session_helper.has_data():
            self.filter_data = self.session_helper.session_filter_data
        else:
            self.filter_data = (
                default_filtering_data if default_filtering_data is not None else {}
            )

        self.remove_old_data()

        if self.user_command in ['filter', 'search'] or self.request.GET.get('page'):
            self.session_helper.add(self.filter_data.copy())

        self._process_commands()

    def _filter_data_contains_filter_by_keys(self) -> bool:
        return (
            self.filter_data and
            any(
                key not in [
                    'search_value', 'csrfmiddlewaretoken', 'filter_key', 'filter_type'
                ]
                for key in self.filter_data
            )
        )

    def _process_commands(self):
        if self.filter_key == self.user_key and self.user_command == 'clear':
            self.session_helper.clear()

    def remove_old_data(self):
        self.session_helper.clean()

    def process_queryset(
            self,
            queryset: SearchQuerySet | FilterQuerySet,
            ignore_search_query: bool = False
    ):
        if self.filter_data.get('search_value') and not ignore_search_query:
            queryset = queryset.search(self.filter_data.get('search_value'))

        if self._filter_data_contains_filter_by_keys():
            queryset = queryset.filter_by_query_dict(self.filter_data)

        return queryset

    def to_context_data(self) -> dict:
        return {
            self.filter_key: {
                'has_data': self.session_helper.has_data(),
                'key': self.filter_key
            }
        }



from __future__ import annotations

from abc import ABC
from typing import Any


class BaseService(ABC):
    class_setup: bool = False

    def __init__(self, obj: Any = None):
        """
            Rules
            - First annotated class attribute must be the target object
            - Must have the same target object name as the class name?
            - Target type must be the same type as the object?
        """

        # Object will be initialized on the __get__ call
        if obj is None:
            return

        self._obj_name: str = list(self.__class__.__annotations__.keys())[0]
        self._obj_type: type = obj.__class__

        setattr(self, self._obj_name, obj)

        # Cascade initializing services
        self._init_sub_services()

    def __init_subclass__(cls):
        super().__init_subclass__()

        if ABC not in cls.__bases__:

            # Typing Does not work properly for services if you override __get__ in the BaseService class.
            # This is a workaround and should be fixed in future versions of the python lsp.
            def __get__(self, instance, owner):
                if instance is None:
                    target: BaseService | Any = owner()
                else:
                    target: BaseService | Any = instance

                # Caches the service and sub services onto the target object.
                if hasattr(target, cls._cache_key()):
                    return getattr(target, cls._cache_key())
                else:
                    instance = cls(target)
                    setattr(target, cls._cache_key(), instance)
                    return instance

            setattr(cls, '__get__', __get__)

    def _init_sub_services(self):
        for key, value in self.__class__.__annotations__.items():
            if isinstance(getattr(self, key), type) and issubclass(getattr(self, key), BaseService):
                setattr(self, key, getattr(self, key)(self.obj))

    @classmethod
    def _cache_key(cls):
        return f'{cls.__name__}_cache_key'

    @property
    def obj(self) -> Any:
        return getattr(self, self._obj_name)

    @property
    def _obj_is_valid(self) -> bool:
        return isinstance(self.obj, self._obj_type)



from abc import abstractmethod

from django.db.models import QuerySet


class FilterQuerySet(QuerySet):
    @abstractmethod
    def filter_by_query_dict(self, filter_data: dict) -> QuerySet:
        raise NotImplementedError('Must implement filter_by method on queryset.')


class SearchQuerySet(QuerySet):
    @abstractmethod
    def search(self, search_value: str) -> QuerySet:
        raise NotImplementedError('Must implement search method on queryset.')



from django.contrib import admin

from django_spire.notification.sms.models import SmsNotification, SmsTemporaryMedia


@admin.register(SmsNotification)
class SmsNotificationAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'notification', 'to_phone_number'
    )


@admin.register(SmsTemporaryMedia)
class SmsTemporaryMediaAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'name', 'content_type', 'external_url'
    )



from abc import ABC
from datetime import datetime, timedelta


class QuerySetFilterSession(ABC):
    _session_key: str = 'list_filter'

    def __init__(self, request, filter_key: str, timeout: int = 60 * 5):
        self.request = request
        self.filter_key = filter_key
        self.timeout = timeout

        if self._session_key not in self.request.session:
            self.request.session[self._session_key] = {}

    @property
    def session_filter_data(self) -> dict:
        return self.request.session[self._session_key].get(self.filter_key, {})

    def add(self, data: dict):
        data['_timeout'] = (datetime.now() + timedelta(seconds=self.timeout)).timestamp()
        data.pop('csrfmiddlewaretoken', None)
        data.pop('filter_type', None)
        data.pop('filter_key', None)
        self.request.session[self._session_key][self.filter_key] = data
        self.request.session.modified = True

    def has_data(self) -> bool:
        return bool(self.session_filter_data)

    def clean(self) -> None:
        if self.has_data():
            current_timestamp = datetime.now().timestamp()
            timeout = self.session_filter_data.get('_timeout', 0)

            if timeout < current_timestamp:
                self.clear()

    def clear(self) -> None:
        if self.has_data():
            del self.request.session[self._session_key][self.filter_key]
            self.request.session.modified = True
